from enum import Enum


class EstimatorType(str, Enum):
    """List of possible estimator types."""

    ClassifierMixin = "classifier"
    OutlierMixin = "outlier"
    RegressorMixin = "regressor"
    TransformerMixin = "transformer"
    ClusterMixin = "cluster"


class TagType(str, Enum):
    """List of extra tags.

    Description of each tag is available at https://scikit-learn.org/dev/developers/develop.html#estimator-tags.
    """

    allow_nan = "allow_nan"
    array_api_support = "array_api_support"
    binary_only = "binary_only"
    multilabel = "multilabel"
    multioutput = "multioutput"
    multioutput_only = "multioutput_only"
    no_validation = "no_validation"
    non_deterministic = "non_deterministic"
    pairwise = "pairwise"
    preserves_dtype = "preserves_dtype"
    poor_score = "poor_score"
    requires_fit = "requires_fit"
    requires_positive_X = "requires_positive_X"  # noqa: N815
    requires_y = "requires_y"
    requires_positive_y = "requires_positive_y"
    _skip_test = "_skip_test"
    _xfail_checks = "_xfail_checks"
    stateless = "stateless"
    X_types = "X_types"



from collections.abc import Callable
from typing import Concatenate, ParamSpec, TypeVar

from result import Err, Ok, Result
from typer import BadParameter, CallbackParam, Context

from sksmithy._models import EstimatorType
from sksmithy._parsers import check_duplicates, name_parser, params_parser, tags_parser

T = TypeVar("T")
R = TypeVar("R")
PS = ParamSpec("PS")


def _parse_wrapper(
    ctx: Context,
    param: CallbackParam,
    value: T,
    parser: Callable[Concatenate[T, PS], Result[R, str]],
    *args: PS.args,
    **kwargs: PS.kwargs,
) -> tuple[Context, CallbackParam, R]:
    """Wrap a parser to handle 'caching' logic."""
    if not ctx.obj:
        ctx.obj = {}

    if param.name in ctx.obj:
        return ctx, param, ctx.obj[param.name]

    result = parser(value, *args, **kwargs)
    match result:
        case Ok(result_value):
            ctx.obj[param.name] = result_value
            return ctx, param, result_value
        case Err(msg):
            raise BadParameter(msg)


def name_callback(ctx: Context, param: CallbackParam, value: str) -> str:
    """`name` argument callback.

    After parsing `name`, changes the default value of `output_file` argument to `{name.lower()}.py`.
    """
    *_, name = _parse_wrapper(ctx, param, value, name_parser)

    # Change default value of output_file argument
    all_options = ctx.command.params
    output_file_option = next(opt for opt in all_options if opt.name == "output_file")
    output_file_option.default = f"{name.lower()}.py"

    return name


def params_callback(ctx: Context, param: CallbackParam, value: str) -> list[str]:
    """`required-params` and `optional-params` arguments callback."""
    ctx, param, parsed_params = _parse_wrapper(ctx, param, value, params_parser)

    if param.name == "optional_params" and (
        msg := check_duplicates(
            required=ctx.params["required_params"],
            optional=parsed_params,
        )
    ):
        del ctx.obj[param.name]
        raise BadParameter(msg)

    return parsed_params


def tags_callback(ctx: Context, param: CallbackParam, value: str) -> list[str]:
    """`tags` argument callback."""
    *_, parsed_value = _parse_wrapper(ctx, param, value, tags_parser)
    return parsed_value


def estimator_callback(ctx: Context, param: CallbackParam, estimator: EstimatorType) -> str:
    """`estimator_type` argument callback.

    It dynamically modifies the behaviour of the rest of the prompts based on its value.
    """
    if not ctx.obj:
        ctx.obj = {}

    if param.name in ctx.obj:
        return ctx.obj[param.name]

    # !Warning: This unpacking relies on the order of the arguments in the forge command to be in the same order.
    # Is there a better/more robust way of dealing with it?
    linear, predict_proba, decision_function = (
        op for op in ctx.command.params if op.name in {"linear", "predict_proba", "decision_function"}
    )

    match estimator:
        case EstimatorType.ClassifierMixin | EstimatorType.RegressorMixin:
            pass
        case _:
            linear.prompt = False  # type: ignore[attr-defined]
            linear.prompt_required = False  # type: ignore[attr-defined]

    match estimator:
        case EstimatorType.ClassifierMixin | EstimatorType.OutlierMixin:
            pass
        case _:
            predict_proba.prompt = False  # type: ignore[attr-defined]
            predict_proba.prompt_required = False  # type: ignore[attr-defined]

    match estimator:
        case EstimatorType.ClassifierMixin:
            pass
        case _:
            decision_function.prompt = False  # type: ignore[attr-defined]
            decision_function.prompt_required = False  # type: ignore[attr-defined]

    ctx.obj[param.name] = estimator.value

    return estimator.value



from typing import Annotated

from typer import Option

from sksmithy._callbacks import estimator_callback, name_callback, params_callback, tags_callback
from sksmithy._models import EstimatorType
from sksmithy._prompts import (
    PROMPT_DECISION_FUNCTION,
    PROMPT_ESTIMATOR,
    PROMPT_LINEAR,
    PROMPT_NAME,
    PROMPT_OPTIONAL,
    PROMPT_OUTPUT,
    PROMPT_PREDICT_PROBA,
    PROMPT_REQUIRED,
    PROMPT_SAMPLE_WEIGHT,
    PROMPT_TAGS,
)

name_arg = Annotated[
    str,
    Option(
        prompt=PROMPT_NAME,
        prompt_required=False,
        help="[bold green]Name[/bold green] of the estimator",
        callback=name_callback,
    ),
]

estimator_type_arg = Annotated[
    EstimatorType,
    Option(
        prompt=PROMPT_ESTIMATOR,
        prompt_required=False,
        help="[bold green]Estimator type[/bold green]",
        callback=estimator_callback,
    ),
]

required_params_arg = Annotated[
    str,
    Option(
        prompt=PROMPT_REQUIRED,
        help="List of [italic yellow](comma-separated)[/italic yellow] [bold green]required[/bold green] parameters",
        callback=params_callback,
    ),
]

optional_params_arg = Annotated[
    str,
    Option(
        prompt=PROMPT_OPTIONAL,
        help="List of  [italic yellow](comma-separated)[/italic yellow] [bold green]optional[/bold green] parameters",
        callback=params_callback,
    ),
]

sample_weight_arg = Annotated[
    bool,
    Option(
        is_flag=True,
        prompt=PROMPT_SAMPLE_WEIGHT,
        prompt_required=False,
        help="Whether or not `.fit()` supports [bold green]`sample_weight`[/bold green]",
    ),
]

linear_arg = Annotated[
    bool,
    Option(
        is_flag=True,
        prompt=PROMPT_LINEAR,
        help="Whether or not the estimator is [bold green]linear[/bold green]",
    ),
]

predict_proba_arg = Annotated[
    bool,
    Option(
        is_flag=True,
        prompt=PROMPT_PREDICT_PROBA,
        help="Whether or not the estimator implements [bold green]`predict_proba`[/bold green] method",
    ),
]

decision_function_arg = Annotated[
    bool,
    Option(
        is_flag=True,
        prompt=PROMPT_DECISION_FUNCTION,
        help="Whether or not the estimator implements [bold green]`decision_function`[/bold green] method",
    ),
]

tags_arg = Annotated[
    str,
    Option(
        prompt=PROMPT_TAGS,
        prompt_required=False,
        help="List of optional extra scikit-learn [bold green]tags[/bold green]",
        callback=tags_callback,
    ),
]

output_file_arg = Annotated[
    str,
    Option(
        prompt=PROMPT_OUTPUT,
        help="[bold green]Destination file[/bold green] where to save the boilerplate code",
    ),
]



from typing import Final

PROMPT_NAME: Final[str] = "🐍 How would you like to name the estimator?"
PROMPT_ESTIMATOR: Final[str] = "🎯 Which kind of estimator is it?"
PROMPT_REQUIRED: Final[str] = "📜 Please list the required parameters (comma-separated)"
PROMPT_OPTIONAL: Final[str] = "📑 Please list the optional parameters (comma-separated)"
PROMPT_SAMPLE_WEIGHT: Final[str] = "📶 Does the `.fit()` method support `sample_weight`?"

PROMPT_LINEAR: Final[str] = "📏 Is the estimator linear?"
PROMPT_PREDICT_PROBA: Final[str] = "🎲 Should the estimator implement a `predict_proba` method?"
PROMPT_DECISION_FUNCTION: Final[str] = "❓ Should the estimator implement a `decision_function` method?"
PROMPT_TAGS: Final[str] = (
    "🧪 We are almost there... Is there any tag you want to add? (comma or space separated)\n"
    "To know more about tags, check the documentation at:\n"
    "https://scikit-learn.org/dev/developers/develop.html#estimator-tags"
)

PROMPT_OUTPUT: Final[str] = "📂 Where would you like to save the class?"



import subprocess
from importlib import resources
from pathlib import Path
from typing import Final

from jinja2 import Template

from sksmithy._models import EstimatorType

TEMPLATE_PATH: Final[Path] = Path(str(resources.files("sksmithy") / "template.py.jinja"))


def render_template(
    name: str,
    estimator_type: EstimatorType,
    required: list[str],
    optional: list[str],
    linear: bool = False,
    sample_weight: bool = False,
    predict_proba: bool = False,
    decision_function: bool = False,
    tags: list[str] | None = None,
) -> str:
    """
    Render a template using the provided parameters.

    Parameters
    ----------
    name: The name of the template.
    estimator_type : The type of the estimator.
    required : A list of required parameters.
    optional : A list of optional parameters.
    linear : Whether the estimator is linear or not
    sample_weight : Whether the estimator supports sample weights in `.fit()` or not
    predict_proba : Whether the estimator will have a `.predict_proba()` method or not.
    decision_function : Whether the estimator will have `.decision_function()` method or not.
    tags : A list of scikit-learn extra tags.

    Returns
    -------
    str : The rendered and formatted template as a string.
    """
    values = {
        "name": name,
        "estimator_type": estimator_type.value,
        "mixin": estimator_type.name,
        "required": required,
        "optional": optional,
        "parameters": [*required, *optional],
        "linear": linear,
        "sample_weight": sample_weight,
        "predict_proba": predict_proba,
        "decision_function": decision_function,
        "tags": tags,
    }

    with TEMPLATE_PATH.open(mode="r") as stream:
        template = Template(stream.read()).render(values)

    return subprocess.check_output(["ruff", "format", "-"], input=template, encoding="utf-8")



from keyword import iskeyword

from result import Err, Ok, Result

from sksmithy._models import TagType


def name_parser(name: str | None) -> Result[str, str]:
    """Validate that `name` is a valid python class name."""
    if name:
        if not name.isidentifier():
            msg = f"`{name}` is not a valid python class name!"
            return Err(msg)
        if iskeyword(name):
            msg = f"`{name}` is a python reserved keyword!"
            return Err(msg)
        return Ok(name)
    msg = "Name cannot be empty!"
    return Err(msg)


def params_parser(params: str | None) -> Result[list[str], str]:
    """Parse and validate that `params` contains valid python names."""
    param_list: list[str] = params.split(",") if params else []
    invalid = tuple(p for p in param_list if not p.isidentifier())

    if len(invalid) > 0:
        msg = f"The following parameters are invalid python identifiers: {invalid}"
        return Err(msg)

    if len(set(param_list)) < len(param_list):
        msg = "Found repeated parameters!"
        return Err(msg)

    return Ok(param_list)


def check_duplicates(required: list[str], optional: list[str]) -> str | None:
    """Check that there are not duplicates between required and optional params."""
    duplicated_params = set(required).intersection(set(optional))
    return (
        f"The following parameters are duplicated between required and optional: {duplicated_params}"
        if duplicated_params
        else None
    )


def tags_parser(tags: str) -> Result[list[str], str]:
    """Parse and validate `tags` by comparing with sklearn list."""
    list_tag: list[str] = tags.split(",") if tags else []

    unavailable_tags = tuple(t for t in list_tag if t not in TagType.__members__)
    if len(unavailable_tags):
        msg = (
            f"The following tags are not available: {unavailable_tags}."
            "\nPlease check the official documentation at "
            "https://scikit-learn.org/dev/developers/develop.html#estimator-tags"
            " to know which values are available."
        )

        return Err(msg)

    return Ok(list_tag)



from pathlib import Path

import typer

from sksmithy._arguments import (
    decision_function_arg,
    estimator_type_arg,
    linear_arg,
    name_arg,
    optional_params_arg,
    output_file_arg,
    predict_proba_arg,
    required_params_arg,
    sample_weight_arg,
    tags_arg,
)
from sksmithy._logger import console
from sksmithy._utils import render_template

app = typer.Typer(
    help="Awesome CLI to generate scikit-learn estimator boilerplate code",
    rich_markup_mode="rich",
    rich_help_panel="Customization and Utils",
)


@app.command()
def version() -> None:
    """Display library version."""
    from importlib import metadata

    __version__ = metadata.version("sklearn-smithy")
    console.print(f"sklearn-smithy {__version__}", style="good")


@app.command()
def forge(
    name: name_arg,
    estimator_type: estimator_type_arg,
    required_params: required_params_arg = "",
    optional_params: optional_params_arg = "",
    sample_weight: sample_weight_arg = False,
    linear: linear_arg = False,
    predict_proba: predict_proba_arg = False,
    decision_function: decision_function_arg = False,
    tags: tags_arg = "",
    output_file: output_file_arg = "",
) -> None:
    """Generate a new shiny scikit-learn compatible estimator ✨

    Depending on the estimator type the following additional information could be required:

    * if the estimator is linear (classifier or regression)
    * if the estimator has a `predict_proba` method (classifier or outlier detector)
    * is the estimator has a `decision_function` method (classifier only)

    Finally, the following two questions will be prompt:

    * if the estimator should have tags (To know more about tags, check the dedicated scikit-learn documentation
        at https://scikit-learn.org/dev/developers/develop.html#estimator-tags
    * in which file the class should be saved (default is `f'{name.lower()}.py'`)
    """
    forged_template = render_template(
        name=name,
        estimator_type=estimator_type,
        required=required_params,  # type: ignore[arg-type]  # Callback transforms it into `list[str]`
        optional=optional_params,  # type: ignore[arg-type]  # Callback transforms it into `list[str]`
        linear=linear,
        sample_weight=sample_weight,
        predict_proba=predict_proba,
        decision_function=decision_function,
        tags=tags,  # type: ignore[arg-type]  # Callback transforms it into `list[str]`
    )

    destination_file = Path(output_file)
    destination_file.parent.mkdir(parents=True, exist_ok=True)

    with destination_file.open(mode="w") as destination:
        destination.write(forged_template)

    console.print(f"Template forged at {destination_file}", style="good")


if __name__ == "__main__":
    app()
